///////////////////////////////////////////////////////////////////////////////
// ketchup/message_processor.hpp
//
//  Version 0.904
//  Copyright 2005 MB@unknown.jp.
//
// This file is a part of the *ketchup*.
// The use and distribution terms for this software are covered by the
// Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
// which can be found in the file CPL.TXT at the root of this distribution.
// By using this software in any fashion, you are agreeing to be bound by
// the terms of this license. You must not remove this notice, or
// any other, from this software.
//
//  Revision History
//   05 May 2005 Initial version

///////////////////////////////////////////////////////////////////////////////
// Concepts:
//
//  MessageMap is...
//   a MPL Sequence of IfStatement.
//

#ifndef KETCHUP_MESSAGE_PROCESSOR_H
#define KETCHUP_MESSAGE_PROCESSOR_H

#ifdef _MSC_VER
#pragma once
#endif

#include <boost/call_traits.hpp>
#include <boost/concept_check.hpp>
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/mpl/apply.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/eval_if.hpp>
#include <boost/mpl/is_sequence.hpp>
#include <boost/mpl/placeholders.hpp>
#include <boost/spirit/fusion/sequence/tuple.hpp>
#include <boost/spirit/fusion/sequence/get.hpp>

#include "if_statement_seq.hpp"

namespace ketchup {

	namespace detail {

	template< class Derived >
	struct pwm_args
	{
		typedef boost::fusion::tuple<
			Derived&, HWND, UINT, WPARAM, LPARAM, LRESULT&, DWORD, BOOL&
		> type;
	};

	template< class Derived >
	struct pwm_args_reference
	{
		typedef typename pwm_args<Derived>::type params_t;
		typedef typename boost::call_traits<params_t>::reference type;
	};

	template< class Derived >
	struct pwm_args_param_type
	{
		typedef typename pwm_args<Derived>::type params_t;
		typedef typename boost::call_traits<params_t>::param_type type;
	};


	///////////////////////////////////////////////////////////////////////////////
	// message_processor_base
	//
	template< class Derived, class HasCrackedHandlers >
	struct message_processor_base
	{
		typedef typename pwm_args<Derived>::type pwm_args_t;
		typedef typename pwm_args_reference<Derived>::type pwm_args_t_reference;
		typedef typename pwm_args_param_type<Derived>::type pwm_args_t_param_type;

		struct final_expr
		{
			typedef BOOL type;
			static type evaluate(pwm_args_t_param_type xs)
			{
				return FALSE;
			}
		};

		struct base
		{
			template< class MessageMap >
			BOOL process_window_message(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT& lResult, DWORD dwMsgMapID)
			{
				Derived& derived = *static_cast<Derived*>(this);
				BOOL bHandled = TRUE;
				pwm_args_t xs(derived, hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID, bHandled);

				typedef typename if_statement_seq<MessageMap, pwm_args_t_reference, final_expr>::type body;
				return body::evaluate(xs);
			};

			typedef base type;
		};	

		struct base_for_cracked_handlers
		{
			bool b_msg_handled;
			bool is_msg_handled() { return b_msg_handled; }
			void set_msg_handled(bool handled) { b_msg_handled = handled; }

			template< class MessageMap >
			BOOL process_window_message(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT& lResult, DWORD dwMsgMapID)
			{
				Derived& derived = *static_cast<Derived*>(this);
				BOOL bHandled = TRUE;
				pwm_args_t xs(derived, hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID, bHandled);

				bool old = derived.is_msg_handled();	// for recursive calls
				typedef typename if_statement_seq<MessageMap, pwm_args_t_reference, final_expr>::type body;
				BOOL ret = body::evaluate(xs);
   			derived.set_msg_handled(old);
				return ret;
			};

			typedef base_for_cracked_handlers type;
		};

		typedef typename boost::mpl::eval_if<HasCrackedHandlers,
			base_for_cracked_handlers,
			base
		>::type type;
	};

	} // namespace detail 


///////////////////////////////////////////////////////////////////////////////
// message_processor
//
	template< class Derived, bool has_cracked_handlers = false >
	struct message_processor :
		detail::message_processor_base< Derived, boost::mpl::bool_<has_cracked_handlers> >::type
	{
	protected:
		typedef typename detail::pwm_args<Derived>::type pwm_args_t;
		typedef typename detail::pwm_args_reference<Derived>::type pwm_args_t_reference;
		typedef typename detail::pwm_args_param_type<Derived>::type	pwm_args_t_param_type;

	///////////////////////////////////////////////////////////////////////////////
	// shortcuts
	//
	static Derived&	get_derived(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<0>(xs);
	}

	static HWND get_hWnd(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<1>(xs);
	}

	static UINT get_uMsg(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<2>(xs);
	}

	static WPARAM get_wParam(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<3>(xs);
	}

	static LPARAM get_lParam(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<4>(xs);
	}

	static LRESULT& get_lResult(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<5>(xs);
	}

	static DWORD get_dwMsgMapID(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<6>(xs);
	}

	static BOOL& get_bHandled(pwm_args_t_param_type xs)
	{
		return boost::fusion::get<7>(xs);
	}

	///////////////////////////////////////////////////////////////////////////////
	// IfStatements of MessageMap
	//
	// Note: Default parameters in nested templates will be ignored by VC7.1.
	//
	struct message_handler_base
	{
		struct returned_expression_type
		{
			static BOOL evaluate(pwm_args_t_param_type xs)
			{
				boost::ignore_unused_variable_warning(xs);
				return TRUE;
			}
			typedef BOOL type;
		};
	};

	struct returned_expr
	{
		static BOOL evaluate(pwm_args_t_param_type xs)
		{
			boost::ignore_unused_variable_warning(xs);
			return TRUE;
		}
		typedef BOOL type;
	};

	///////////////////////////////////////////////////////////////////////////////
	// empty_handler
	//
	struct empty_handler_cond_expr
	{
		static bool evaluate(pwm_args_t_reference xs)
		{
			boost::ignore_unused_variable_warning(xs);
			return false;
		}
		typedef bool type;
	};

	typedef typename gen_if_statement<empty_handler_cond_expr, returned_expr>::type empty_handler;

	///////////////////////////////////////////////////////////////////////////////
	// message_handler
	//
	typedef LRESULT (Derived::*message_handler_type)(UINT, WPARAM, LPARAM, BOOL&);

	template< UINT msg, message_handler_type func >
	struct message_handler_cond_expr
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == msg) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(get_uMsg(xs), get_wParam(xs), get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	typedef typename message_handler<

	///////////////////////////////////////////////////////////////////////////////
	// message_range_handler
	//
	template< UINT msgFirst, UINT msgLast, message_handler_type func >
	struct message_range_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) >= msgFirst && get_uMsg(xs) <= msgLast) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(get_uMsg(xs), get_wParam(xs), get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_handler
	//
	typedef LRESULT (Derived::*command_handler_type)(WORD, WORD, HWND, BOOL&);

	template< WORD id, WORD code, command_handler_type func >
	struct command_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == WM_COMMAND && id == LOWORD(get_wParam(xs)) && code == HIWORD(get_wParam(xs))) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(HIWORD(get_wParam(xs)), LOWORD(get_wParam(xs)), (HWND)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_id_handler
	//
	template< WORD id, command_handler_type func >
	struct command_id_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if(get_uMsg(xs) == WM_COMMAND && id == LOWORD(get_wParam(xs))) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(HIWORD(get_wParam(xs)), LOWORD(get_wParam(xs)), (HWND)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_code_handler
	//
	template< WORD code, command_handler_type func >
	struct command_code_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if(get_uMsg(xs) == WM_COMMAND && id == HIWORD(get_wParam(xs))) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(HIWORD(get_wParam(xs)), LOWORD(get_wParam(xs)), (HWND)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_range_handler
	//
	template< WORD idFirst, WORD idLast, command_handler_type func >
	struct command_range_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if(get_uMsg(xs) == WM_COMMAND && LOWORD(get_wParam(xs)) >= idFirst && LOWORD(get_wParam(xs)) <= idLast) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(HIWORD(get_wParam(xs)), LOWORD(get_wParam(xs)), (HWND)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_range_code_handler
	//
	template< WORD idFirst, WORD idLast, WORD code, command_handler_type func >
	struct command_range_code_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if(get_uMsg(xs) == WM_COMMAND && code == HIWORD(get_wParam(xs)) && LOWORD(get_wParam(xs)) >= idFirst && LOWORD(get_wParam(xs)) <= idLast) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)(HIWORD(get_wParam(xs)), LOWORD(get_wParam(xs)), (HWND)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_handler
	//
	typedef LRESULT (Derived::*notify_handler_type)(int, LPNMHDR, BOOL&);

	template< UINT id, UINT cd, notify_handler_type func >
	struct notify_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == WM_NOTIFY && id == ((LPNMHDR)get_lParam(xs))->idFrom && cd == ((LPNMHDR)get_lParam(xs))->code) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)((int)get_wParam(xs), (LPNMHDR)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_id_handler
	//
	template< UINT id, notify_handler_type func >
	struct notify_id_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == WM_NOTIFY && id == ((LPNMHDR)get_lParam(xs))->idFrom) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)((int)get_wParam(xs), (LPNMHDR)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_code_handler
	//
	template< UINT cd, notify_handler_type func >
	struct notify_code_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == WM_NOTIFY && cd == ((LPNMHDR)get_lParam(xs))->code) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)((int)get_wParam(xs), (LPNMHDR)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_range_handler
	//
	template< UINT idFirst, UINT idLast, notify_handler_type func >
	struct notify_range_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == WM_NOTIFY && ((LPNMHDR)get_lParam(xs))->idFrom >= idFirst && ((LPNMHDR)get_lParam(xs))->idFrom <= idLast) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)((int)get_wParam(xs), (LPNMHDR)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_range_code_handler
	//
	template< UINT idFirst, UINT idLast, UINT cd, notify_handler_type func >
	struct notify_range_code_handler : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_uMsg(xs) == WM_NOTIFY &&  cd == ((LPNMHDR)get_lParam(xs))->code && ((LPNMHDR)get_lParam(xs))->idFrom >= idFirst && ((LPNMHDR)get_lParam(xs))->idFrom <= idLast) {
					get_bHandled(xs) = TRUE;
					get_lResult(xs) = (get_derived(xs).*func)((int)get_wParam(xs), (LPNMHDR)get_lParam(xs), get_bHandled(xs));
					if (get_bHandled(xs))
						return true;
				}
				return false;
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// chain_msg_map
	//
	//  Note: to avoid ::type syntax
	//
	template< class theChainClassOrMessageMap >
	struct chain_msg_map_base
	{
		template< class MessageMap >
		struct chain_if_statement : message_handler_base
		{
			struct condition_expression_type
			{
				static bool evaluate(pwm_args_t_reference xs)
				{
					return get_derived(xs).process_window_message<MessageMap>(get_hWnd(xs), get_uMsg(xs), get_wParam(xs), get_lParam(xs), get_lResult(xs), get_dwMsgMapID(xs));
				}
			};

			typedef chain_if_statement type;
		};

		template< class theChainClass >
		struct chain_if_statement_Atl : message_handler_base
		{
			struct condition_expression_type
			{
				static bool evaluate(pwm_args_t_reference xs)
				{
					if (get_derived(xs).theChainClass::ProcessWindowMessage(get_hWnd(xs), get_uMsg(xs), get_wParam(xs), get_lParam(xs), get_lResult(xs), get_dwMsgMapID(xs)))
						return true;
					return false;
				}
			};

			typedef chain_if_statement_Atl type;
		};

		typedef typename boost::mpl::eval_if<boost::mpl::is_sequence<theChainClassOrMessageMap>,
			chain_if_statement<boost::mpl::_>,
			chain_if_statement_Atl<boost::mpl::_>
		>::type if_statement;

		typedef typename boost::mpl::apply<if_statement, theChainClassOrMessageMap>::type type;
	};

	template< class theChainClassOrMessageMap >
	struct chain_msg_map : chain_msg_map_base<theChainClassOrMessageMap>::type
	{ };

	///////////////////////////////////////////////////////////////////////////////
	// alt_msg_map
	//
	template< DWORD msgMapID, class MessageMap >
	struct alt_msg_map : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				if (get_dwMsgMapID(xs) == msgMapID) {
					return get_derived(xs).process_window_message<MessageMap>(get_hWnd(xs), get_uMsg(xs), get_wParam(xs), get_lParam(xs), get_lResult(xs), get_dwMsgMapID(xs));
				}
				else
					return false; // pass to other alternates
			}
		};
	};

	///////////////////////////////////////////////////////////////////////////////
	// cracked_handlers
	//

//#	include "detail/cracked_handlers.ipp"
	template<
		void (Derived::*func)()
	>
	struct msg_wm_close : message_handler_base
	{
		struct condition_expression_type
		{
			static bool evaluate(pwm_args_t_reference xs)
			{
				BOOST_STATIC_ASSERT(has_cracked_handlers);

				if (get_uMsg(xs) == WM_CLOSE) {
					get_derived(xs).set_msg_handled(true);
					(get_derived(xs).*func)();
					get_lResult(xs) = 0;
					if (get_derived(xs).is_msg_handled())
						return true;
				}
				return false;
			}
		};
	};


}; // struct message_processor

} // namespace ketchup

///////////////////////////////////////////////////////////////////////////////
// KETCHUP_CHAIN_MSG_MAP for ATL's BEGIN_MSG_MAP
// 
#define KETCHUP_CHAIN_MSG_MAP(MessageMap) {\
	if (process_window_message<MessageMap>(hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID))\
		return TRUE;\
}

#endif
