///////////////////////////////////////////////////////////////////////////////
// if_statement_seq.hpp
//
// Concepts:
//  IfStatementSeq is...
//   a MPL Sequence of IfStatement.
//
//  Statement is...
//   a type which has two static member functions...
//
//    Statement::type compile(Args) {
//     if you want to return from the body, return true,
//     else you want to continue, return false.
//    }
//
//
//
//  Args is...
//   AnyType.
//
//  if_statement_seq<StatementSequence, Argument> is...
//   a metafunction which returns Statement
//   generated from StatementSequence.
//
// Note:
//  big_if_statement can't touch Argument type,
//  so you should choose whether referenced or not.
//

#ifndef KETCHUP_BIG_IF_STATEMENT_H
#define KETCHUP_BIG_IF_STATEMENT_H

#ifdef _MSC_VER
#pragma once
#endif

#include <boost/concept_check.hpp>
#include <boost/mpl/reverse_fold.hpp>

namespace ketchup {

	namespace detail {

	///////////////////////////////////////////////////////////////////////////////
	// Note:
	//  If StatementSequence has the same Statement, VC7.1 duplicates
	//  the static member function, Statement::compile (LNK1179 error).
	//  I can't understand the problem,
	//  but VC7.1 seems to hate static member functions.
	//  So I define fold_op_workaround and...
	//
	// compile_statement
	//
	template< class ConditionExpression, class Args >
	typename ConditionExpression::type
	compile_condition_expression(Args xs)
	{
		return ConditoinExpression::compile(xs);
	}

	template< class ReturnStatement, class Args >
	typename ReturnStatement::type
	compile_return_statement(Args xs)
	{
		return ReturnStatement::compile(xs);
	}

	} // namespace detail


///////////////////////////////////////////////////////////////////////////////
// helper metafunctions
//
template< class IfStatement >
struct get_return_statement
{
	typedef typename IfStatement::return_statement_type type;
}

template< class IfStatement >
struct get_condition_expression
{
	typedef typename IfStatement::condition_expression_type type;
}

/*
template< class Args, class ConditionExpression, class ReturnStatement >
struct if_condition_return_statement
{
	typedef typename ReturnStatement::type type;
	static type compile(Args xs)
	{
		if (ConditionExpression::compile(xs))
			return ReturnStatement::compile(xs);
		else
			return State::compile(x);		
	}
};*/

///////////////////////////////////////////////////////////////////////////////
// if_statement_seq is...
//  a Statement.
//
// Args is...
//  AnyType.
// IfStatements is...
//  a MPL Sequence of IfStatement.
// FinalReturnStatement is...
//  a Statement.
//
template<
	class Args, class IfStatements, class FinalReturnStatement
>
struct gen_if_statement_seq_base
{
	typedef typename FinalReturnStatement::type result_type;

	// The placeholders say mpl_assert_not_arg, so I define
	struct fold_op_workaround
	{
		template< class State, class IfStatement >
		struct apply
		{
			typedef typename get_condition_expression<IfStatement>::type cond_exp_t;
			typedef typename get_return_statement<IfStatement>::type ret_statement_t;

			static result_type compile(Args xs)
			{
				if (detail::compile_condition_expression<cond_exp_t, Args>(xs))
					return detail::compile_return_statement<ret_statement_t, Args>(xs);
				else
					return State::compile(x);
			}

			typedef apply type;
		};
	};

	typedef typename boost::mpl::reverse_fold<IfStatements,
		FinalReturnStatement,
		// fold_op<boost::mpl::_1, boost::mpl::_2> // makes mpl_assert_not_arg
		fold_op_workaround
	>::type type;
};

template<
	class Args, class IfStatements, class ReturnStatement
>
struct if_statement_seq : gen_if_statement_base<Args, IfStatements, ReturnStatement>::type
{
	typedef typename ReturnStatement::type type;
};



} // namespace ketchup

#endif
